from distutils.core import setup


setup(
    name="ase-decaf",
    version='0.1.0',
    packages=['decaf'],
    license='GNU General Public License v3.0',
    long_description=open('README.txt').read(),
    author='Sencer Selcuk',
    author_email='sselcuk@bnl.gov'
)
